from __future__ import annotations

from fractions import Fraction
from functools import partial
from math import gcd
from typing import SupportsFloat

from vskernels import BicubicDidee, Catrom
from vstools import (
    CustomError, CustomValueError, Dar, FieldBased, FieldBasedT, FuncExceptT, Region, core, depth, fallback, get_prop,
    get_w, mod2, mod4, vs
)

from .helpers import check_ivtc_pattern

__all__ = [
    'seek_cycle',

    'check_patterns',

    'PARser'
]


def seek_cycle(clip: vs.VideoNode, write_props: bool = True, scale: int = -1) -> vs.VideoNode:
    """
    Purely visual tool to view telecining cycles.

    .. warning::
        | This is purely a visual tool and has no matching parameters!
        | Just use `Wobbly <https://github.com/dubhater/Wobbly>`_ instead if you need that.

    Displays the current frame, two previous and future frames,
    and whether they are combed or not.

    ``P`` indicates a progressive frame,
    and ``C`` a combed frame.

    Dependencies:

    * `VapourSynth-TDeintMod <https://github.com/HomeOfVapourSynthEvolution/VapourSynth-TDeintMod>`_

    :param clip:            Clip to process.
    :param write_props:     Write props on frames. Disabling this will also speed up the function.
    :param scale:           Integer scaling of all clips. Must be to the power of 2.

    :return:                Viewing UI for standard telecining cycles.

    :raises ValueError:     `scale` is a value that is not to the power of 2.
    """

    if (scale & (scale - 1) != 0) and scale != 0 and scale != -1:
        raise CustomValueError("'scale' must be a power of 2!", seek_cycle)

    # TODO: 60i checks and flags somehow? false positives gonna be a pain though
    def check_combed(n: int, f: vs.VideoFrame, clip: vs.VideoNode) -> vs.VideoNode:
        return clip.text.Text("C" if get_prop(f, "_Combed", int) else "P", 7)

    # Scaling of the main clip
    scale = 1 if scale == -1 else 2 ** scale

    height = clip.height * scale
    width = get_w(height, clip.width / clip.height)

    clip = clip.tdm.IsCombed()
    clip = Catrom().scale(clip, width, height)

    # Downscaling for the cycle clips
    clip_down = BicubicDidee().scale(clip, mod2(width / 4), mod2(height / 4))
    if write_props:
        clip_down = core.std.FrameEval(clip_down, partial(check_combed, clip=clip_down), clip_down).text.FrameNum(2)
    blank_frame = clip_down.std.BlankClip(length=1, color=[0] * 3)

    pad_c = clip_down.text.Text("Current", 8) if write_props else clip_down
    pad_a, pad_b = blank_frame * 2 + clip_down[:-2], blank_frame + clip_down[:-1]
    pad_d, pad_e = clip_down[1:] + blank_frame, clip_down[2:] + blank_frame * 2

    # Cycling
    cycle_clips = [pad_a, pad_b, pad_c, pad_d, pad_e]
    pad_x = [pad_a.std.BlankClip(mod4(pad_a.width / 15))] * 4
    cycle = cycle_clips + pad_x  # no shot this can't be done way cleaner
    cycle[::2], cycle[1::2] = cycle_clips, pad_x

    # Final stacking
    stack_abcde = core.std.StackHorizontal(cycle)

    vert_pad = stack_abcde.std.BlankClip(height=mod2(stack_abcde.height / 5))
    horz_pad = clip.std.BlankClip(mod2((stack_abcde.width - clip.width) / 2))

    stack = core.std.StackHorizontal([horz_pad, clip, horz_pad])
    return core.std.StackVertical([vert_pad, stack, vert_pad, stack_abcde])


def check_patterns(clip: vs.VideoNode, tff: bool | FieldBasedT | None = None) -> int:
    """
    Naïve function that iterates over a given clip and tries out every simple 3:2 IVTC pattern.

    This function will return the best pattern value that didn't result in any combing.
    If all of them resulted in combing, it will raise an error.

    Note that the clip length may seem off because I grab every fourth frame of a clip.
    This should make processing faster, and it will still find combed frames.

    This function should only be used for rudimentary testing.
    If I see it in any proper scripts, heads will roll.

    Dependencies:

    * `VapourSynth-TDeintMod <https://github.com/HomeOfVapourSynthEvolution/VapourSynth-TDeintMod>`_

    :param clip:                    Clip to process.
    :param tff:                     Top-field-first. `False` sets it to Bottom-Field-First.
                                    If None, get the field order from the _FieldBased prop.

    :return:                        Integer representing the best pattern.

    :raises TopFieldFirstError:     No automatic ``tff`` can be determined.
    :raises StopIteration:          No pattern resulted in a clean match.
    """

    clip = FieldBased.ensure_presence(clip, tff, check_patterns)

    clip = depth(clip, 8)

    pattern = -1

    for n in [int(n) for n in range(0, 4)]:
        check = check_ivtc_pattern(clip, n)

        if check:
            pattern = n
            break

    if pattern == -1:
        raise CustomError[StopIteration](
            'None of the patterns resulted in a clip without combing. '
            'Please try performing proper IVTC on the clip.', check_patterns
        )

    return pattern


def PARser(
    clip: vs.VideoNode, active_area: int,
    dar: Dar | str | Fraction | None = None, height: int | None = None,
    region: Region | str = Region.NTSC,
    return_result: bool = False, func: FuncExceptT | None = None
) -> vs.VideoNode | dict[str, SupportsFloat | tuple[int, int] | str]:
    """
    Calculate SAR (sample aspect ratio) and attach result as frameprops.

    The active area is calculated by subtracting the dirty edges from the left and right side of the clip.
    Only take the darker pixels into account, and make sure you gather them from bright scenes!
    The SAR and anamorphic width/height will be added to the clip as props, but can also be printed.

    List of common valid active areas (note that ``active_area`` only sets the width!):

        * 704x480 (NTSC, MPEG-4)
        * 708x480 (NTSC, SMPTE Rp87-1995)
        * 710.85x486 (NTSC, ITU-R REC.601)
        * 711x480 (NTSC, MPEG-4)

    Make sure you absolutely DO NOT ACTUALLY CROP in the direction you're stretching!
    If you do, the end result will be off, and the aspect ratio will be wrong! Just leave them alone.
    It's okay to crop the top/bottom when going widescreen, and left/right when going fullscreen.

    If you absolutely *must* crop the pixels away, consider letting the video decoder handle that during playback.
    See: `display-window (x265) <https://x265.readthedocs.io/en/master/cli.html#cmdoption-display-window>`_

    It's not recommended to stretch to the final resolution after calculating the SAR.
    Instead, set the SAR in your encoder's settings and encode your video as anamorphic.
    This will result in the most accurate final image without introducing compounding resampling artefacting
    (don't worry, plenty programs still support anamorphic video).

    Core idea originated from a `private gist <https://gist.github.com/wiwaz/40883bae396bef5eb9fc99d4de2377ec>`_
    and was heavily modified by LightArrowsEXE.

    For more information, I highly recommend reading
    `this blogpost <https://web.archive.org/web/20140218044518/http://lipas.uwasa.fi/~f76998/video/conversion/>`_.

    :param clip:                Input clip.
    :param active_area:         Width you would end up with post-cropping.
                                Only take into account darker messed up edges!
    :param dar:                 Display Aspect Ratio. Refers to the analog television aspect ratio.
                                Must be a :py:attr:`vstools.Dar` enum, a string representing a Dar value,
                                or a Fraction object containing a user-defined DAR.
                                If None, automatically guesses DAR based on the SAR props.
                                Default: assume based on current SAR properties.
    :param height:              Height override. If None, auto-select based on region.
                                This is not particularly useful unless you want to set it to 486p
                                (to use with for example a 1920x1080 -> 864x486 -> 864x480 downscaled + cropped DVD
                                where the studio did not properly account for anamorphic resolutions)
                                or need to deal with ITU-R REC.601 video.
                                Default: input clip's height.
    :param region:              Analog television region. Must be either NTSC or PAL.
                                Must be a :py:attr:`vstools.Region` enum  or string representing a Region value.
                                Default: :py:attr:`vstools.Region.NTSC`.
    :param return_result:       Return the results as a dict. Default: False.

    :return:                    Clip with corrected SAR props and anamorphic width/height prop,
                                or a dictionary with all the results.

    :raises FramePropError:     DAR is None and no SAR props are set on the input clip.
    :raises ValueError:         Invalid :py:attr:`vstools.Dar` is passed.
    :raises ValueError:         Invalid :py:attr:`vstools.Region` is passed.
    """

    func = fallback(func, PARser)

    match dar:
        case Fraction(): new_dar = dar.numerator, dar.denominator
        case Dar.WIDE: new_dar = 16, 9
        case Dar.FULL: new_dar = 4, 3
        case Dar.SQUARE: return clip.std.SetFrameProps(_SARDen=1, _SARNum=1)
        case None: return PARser(clip, active_area, Dar.from_video(clip), height, region, return_result, func)
        case _: raise CustomValueError('Invalid DAR passed! Must be in {values} or None!', func, values=iter(Dar))

    props = dict[str, SupportsFloat | tuple[int, int] | str](dar=new_dar)

    if height is None:
        match region:
            case Region.NTSC: height = 480
            case Region.PAL: height = 576
            case _: raise CustomValueError('Invalid Region passed! Must be in {values}!', func, values=iter(Region))

    sar = new_dar[0] * height, new_dar[1] * active_area
    sargcd = gcd(sar[0], sar[1])

    sarden = sar[0] // sargcd
    sarnum = sar[1] // sargcd

    props.update(_SARDen=sarden, _SARNum=sarnum)

    match dar:
        case Dar.WIDE: props.update(amorph_width=clip.width * (sarden / sarnum))
        case Dar.FULL: props.update(amorph_height=clip.height * (sarnum / sarden))
        # TODO: autoguess which to return based on the sarnum maybe?
        case _: props.update(
            amorph__note="Use your best judgment to pick one!",
            amorph_width=clip.width * (sarden / sarnum),
            amorph_height=clip.height * (sarnum / sarden)
        )

    if return_result:
        return props

    return clip.std.SetFrameProps(**props)



from __future__ import annotations

import gc
import os
import sys
from pathlib import Path
from typing import Any

from vstools import CustomTypeError, FieldBased, FieldBasedT, core, get_render_progress, vs

__all__ = [
    'sivtc',
    'tivtc_vfr'
]


def sivtc(clip: vs.VideoNode, pattern: int = 0, tff: bool | FieldBasedT = True, decimate: bool = True) -> vs.VideoNode:
    """
    Simplest form of a fieldmatching function.

    This is essentially a stripped-down JIVTC offering JUST the basic fieldmatching and decimation part.
    As such, you may need to combine multiple instances if patterns change throughout the clip.

    :param clip:        Clip to process.
    :param pattern:     First frame of any clean-combed-combed-clean-clean sequence.
    :param tff:         Top-Field-First.
    :param decimate:    Drop a frame every 5 frames to get down to 24000/1001.

    :return:            IVTC'd clip.
    """

    pattern = pattern % 5

    defivtc = core.std.SeparateFields(clip, tff=FieldBased.from_param(tff).field).std.DoubleWeave()
    selectlist = [[0, 3, 6, 8], [0, 2, 5, 8], [0, 2, 4, 7], [2, 4, 6, 9], [1, 4, 6, 8]]
    dec = core.std.SelectEvery(defivtc, 10, selectlist[pattern]) if decimate else defivtc
    return dec.std.SetFieldBased(0).std.SetFrameProp(prop='SIVTC_pattern', intval=pattern)


main_file = os.path.realpath(sys.argv[0]) if sys.argv[0] else None
main_file = f"{os.path.splitext(os.path.basename(str(main_file)))[0]}_"
main_file = "{yourScriptName}_" if main_file in ("__main___", "setup_") else main_file


def tivtc_vfr(
    clip: vs.VideoNode,
    tfm_in: Path | str = f".ivtc/{main_file}matches.txt",
    tdec_in: Path | str = f".ivtc/{main_file}metrics.txt",
    timecodes_out: Path | str = f".ivtc/{main_file}timecodes.txt",
    decimate: int | bool = True,
    tfm_args: dict[str, Any] | None = None,
    tdecimate_args: dict[str, Any] | None = None
) -> vs.VideoNode:
    """
    Perform TFM and TDecimate on a clip that is supposed to be VFR.

    Includes automatic generation of a metrics/matches/timecodes txt file.

    | This function took *heavy* inspiration from atomchtools.TIVTC_VFR,
    | and is basically an improved rewrite on the concept.

    .. warning::
        | When calculating the matches and metrics for the first time, your previewer may error out!
        | To fix this, simply refresh your previewer. If it still doesn't work, open the ``.ivtc`` directory
        | and check if the files are **0kb**. If they are, **delete them** and run the function again.
        | You may need to restart your previewer entirely for it to work!

    Dependencies:

    * `TIVTC <https://github.com/dubhater/vapoursynth-tivtc>`_

    :param clip:                Clip to process.
    :param tfmIn:               File location for TFM's matches analysis.
                                By default it will be written to ``.ivtc/{yourScriptName}_matches.txt``.
    :param tdecIn:              File location for TDecimate's metrics analysis.
                                By default it will be written to ``.ivtc/{yourScriptName}_metrics.txt``.
    :param timecodes_out:       File location for TDecimate's timecodes analysis.
                                By default it will be written to ``.ivtc/{yourScriptName}_timecodes.txt``.
    :param decimate:            Perform TDecimate on the clip if true, else returns TFM'd clip only.
                                Set to -1 to use TDecimate without TFM.
    :param tfm_args:            Additional arguments to pass to TFM.
    :param tdecimate_args:      Additional arguments to pass to TDecimate.

    :return:                    IVTC'd VFR clip with external timecode/matches/metrics txt files.

    :raises TypeError:          Invalid ``decimate`` argument is passed.
    """

    decimate = int(decimate)

    if decimate not in {-1, 0, 1}:
        raise CustomTypeError(
            "Invalid 'decimate' argument. Must be True/False, their integer values, or -1!", tivtc_vfr
        )

    tfm_args, tdecimate_args = tfm_args or {}, tdecimate_args or {}

    tfm_f = tdec_f = timecodes_f = Path()

    def _set_paths() -> None:
        nonlocal tfm_f, tdec_f, timecodes_f
        tfm_f = Path(tfm_in).resolve().absolute()
        tdec_f = Path(tdec_in).resolve().absolute()
        timecodes_f = Path(timecodes_out).resolve().absolute()

    _set_paths()

    for p in (tfm_f, tdec_f, timecodes_f):
        p.parent.mkdir(parents=True, exist_ok=True)
        if p.exists() and p.stat().st_size == 0:
            p.unlink(True)

    if not (tfm_f.exists() and tdec_f.exists()):
        ivtc_clip = clip.tivtc.TFM(output=str(tfm_f), **tfm_args)
        ivtc_clip = ivtc_clip.tivtc.TDecimate(mode=4, output=str(tdec_f), **tdecimate_args)

        with get_render_progress() as pr:
            task = pr.add_task("calculating matches and metrics...", total=ivtc_clip.num_frames)

            for _ in ivtc_clip.frames(close=True):
                pr.update(task, advance=1)

        ivtc_clip = None  # type: ignore
        del ivtc_clip

        gc.collect(0)
        gc.collect(1)
        gc.collect(2)

        _set_paths()

    if decimate != -1:
        clip = clip.tivtc.TFM(**tfm_args, input=str(tfm_f))

    if decimate == 0:
        return clip

    return clip.tivtc.TDecimate(
        **tdecimate_args, mode=5, hybrid=2, vfrDec=1, input=str(tdec_f), tfmIn=str(tfm_f), mkvOut=str(timecodes_f)
    )



from __future__ import annotations

from vstools import get_prop, get_render_progress, vs

__all__ = [
    'check_ivtc_pattern'
]


def check_ivtc_pattern(clip: vs.VideoNode, pattern: int = 0) -> bool:
    """:py:func:`vsdeinterlace.utils.check_patterns` rendering behaviour."""

    from .ivtc import sivtc

    clip = sivtc(clip, pattern).tdm.IsCombed()

    p = get_render_progress()
    task = p.add_task(f"Checking pattern {pattern}...", total=clip.num_frames)

    for f in clip[::4].frames(close=True):
        if get_prop(f, '_Combed', int):
            print(f"check_patterns: 'Combing found with pattern {pattern}!'")
            return False

        p.update(task, advance=1)

    print(f"check_patterns: 'Clean clip found with pattern {pattern}!'")
    return True



from __future__ import annotations

from vstools import (
    CustomOverflowError, DependencyNotFoundError, FieldBased, FieldBasedT, check_variable_format, core, depth,
    get_depth, get_neutral_value, scale_value, vs
)

__all__ = [
    'fix_telecined_fades',

    'vinverse'
]


def fix_telecined_fades(clip: vs.VideoNode, tff: bool | FieldBasedT | None = None, cuda: bool = False) -> vs.VideoNode:
    """
    Give a mathematically perfect solution to fades made *after* telecining (which made perfect IVTC impossible).

    This is an improved version of the Fix-Telecined-Fades plugin
    that deals with overshoot/undershoot by adding a check.

    Make sure to run this *after* IVTC/deinterlacing!

    If the value surpases thr * original value, it will not affect any pixels in that frame
    to avoid it damaging frames it shouldn't need to. This helps a lot with orphan fields as well,
    which would otherwise create massive swings in values, sometimes messing up the fade fixing.

    .. warning::
        | If you pass your own float clip, you'll want to make sure to properly dither it down after.
        | If you don't do this, you'll run into some serious issues!

    :param clip:                            Clip to process.
    :param tff:                             Top-field-first. `False` sets it to Bottom-Field-First.
                                            If `None`, get the field order from the _FieldBased prop.
    :param cuda:                            Use cupy for certain calculations. `False` uses numpy instead.

    :return:                                Clip with fades (and only fades) accurately deinterlaced.

    :raises UndefinedFieldBasedError:       No automatic ``tff`` can be determined.
    """

    bits = get_depth(clip)

    clip = FieldBased.ensure_presence(clip, tff, fix_telecined_fades)

    fields = clip.std.Limiter().std.SeparateFields()

    if cuda:
        try:
            from stgfunc import mean_plane_value
        except ModuleNotFoundError:
            raise DependencyNotFoundError(fix_telecined_fades, 'stgfunc', reason='cuda=True')

        avg_props_clip = mean_plane_value(fields, [0.0, 1.0], cuda=True, prop='psmAvg', single_out=True, planes=0)
    else:
        avg_props_clip = fields.psm.PlaneAverage(value_exclude=[0.0, 1.0])  # type: ignore

    def props_bodge(f: list[vs.VideoFrame], n: int) -> vs.VideoFrame:
        fout = f[0].copy()
        fout.props.update(ftAvg=f[1].props.psmAvg, fbAvg=f[2].props.psmAvg)
        return fout

    props_clip = clip.std.ModifyFrame(
        [clip, avg_props_clip[::2], avg_props_clip[1::2]], props_bodge
    )

    output = props_clip.akarin.Expr(
        'Y 2 % BF! BF@ x.fbAvg x.ftAvg ? TAVG! TAVG@ 0 = x x TAVG@ BF@ x.ftAvg x.fbAvg ? + 2 / TAVG@ / * ?'
    )

    return depth(output, bits)


def vinverse(clip: vs.VideoNode, sstr: float = 2.0, amount: int = 128, scale: float = 1.5) -> vs.VideoNode:
    """
    Clean up residual combing after a deinterlacing pass.

    This is Setsugen no ao's implementation, adopted into vsdeinterlace.

    :param clip:        Clip to process.
    :param sstr:        Contrasharpening strength. Increase this if you find
                        the decombing blurs the image a bit too much.
    :param amount:      Maximum difference allowed between the original pixels and adjusted pixels.
                        Scaled to input clip's depth. Set to 255 to effectively disable this.
    :param scale:       Scale amount for vertical sharp * vertical blur.

    :return:            Clip with residual combing largely removed.

    :raises ValueError: ``amount`` is set above 255.
    """

    assert check_variable_format(clip, "vinverse")

    if amount > 255:
        raise CustomOverflowError("'amount' may not be set higher than 255!", vinverse)

    neutral = get_neutral_value(clip)

    # Expression to find combing and separate it from the rest of the clip
    find_combs = clip.akarin.Expr(f'{neutral} n! x x 2 * x[0,-1] x[0,1] + + 4 / - n@ +')

    # Expression to decomb it (creates blending)
    decomb = core.akarin.Expr(
        [find_combs, clip],
        f'{neutral} n! x 2 * x[0,-1] x[0,1] + + 4 / blur! y x blur@ - x n@ - * 0 < n@ x blur@ '
        ' - abs x n@ - abs < x blur@ - n@ + x ? ? - n@ +'
    )

    # Final expression to properly merge it and avoid creating too much damage
    return core.akarin.Expr(
        [clip, decomb],
        f'{neutral} n! {scale_value(amount, 8, clip.format.bits_per_sample)} a! y y y y 2 * y[0,-1]'
        f' y[0,1] + + 4 / - {sstr} * + y - n@ + sdiff! x y - n@ + diff! sdiff@ n@ - diff@ n@ - '
        f'* 0 < sdiff@ n@ - abs diff@ n@ - abs < sdiff@ diff@ ? n@ - {scale} * n@ + sdiff@ n@ '
        '- abs diff@ n@ - abs < sdiff@ diff@ ? ? n@ - + merge! x a@ + merge@ < x a@ + x a@ - '
        'merge@ > x a@ - merge@ ? ?'
    )



from __future__ import annotations

import warnings
from functools import partial

from vsrgtools import repair
from vstools import core, vs

__all__ = [
    'deblend'
]


def deblend(clip: vs.VideoNode, start: int = 0, rep: int | None = None, decimate: bool = True) -> vs.VideoNode:
    """
    Deblending function for blended AABBA patterns.

    .. warning:
        This function's base functionality and settings will be updated in a future version!

    Assuming there's a constant pattern of frames (labeled A, B, C, CD, and DA in this function),
    blending can be fixed by calculating the D frame by getting halves of CD and DA, and using that
    to fix up CD. DA is then dropped because it's a duplicated frame.

    Doing this will result in some of the artifacting being added to the deblended frame,
    but we can mitigate that by repairing the frame with the non-blended frame before it.

    For more information, please refer to `this blogpost by torchlight
    <https://mechaweaponsvidya.wordpress.com/2012/09/13/adventures-in-deblending/>`_.

    :param clip:        Clip to process.
    :param start:       First frame of the pattern (Default: 0).
    :param rep:         Repair mode for the deblended frames, no repair if None (Default: None).
    :param decimate:    Decimate the video after deblending (Default: True).

    :return:            Deblended clip.
    """
    warnings.warn("deblend: 'This function's base functionality and settings "
                  "will be updated in a future version!'", DeprecationWarning)

    blends_a = range(start + 2, clip.num_frames - 1, 5)
    blends_b = range(start + 3, clip.num_frames - 1, 5)
    expr_cd = ["z a 2 / - y x 2 / - +"]

    # Thanks Myaa, motbob and kageru!
    def deblend(n: int, clip: vs.VideoNode, rep: int | None) -> vs.VideoNode:
        if n % 5 in [0, 1, 4]:
            return clip
        else:
            if n in blends_a:
                c, cd, da, a = clip[n - 1], clip[n], clip[n + 1], clip[n + 2]
                debl = core.akarin.Expr([c, cd, da, a], expr_cd)
                return repair(debl, c, rep) if rep else debl
            return clip

    debl = core.std.FrameEval(clip, partial(deblend, clip=clip, rep=rep))
    return core.std.DeleteFrames(debl, blends_b).std.AssumeFPS(fpsnum=24000, fpsden=1001) if decimate else debl
